/*
 * Copyright (C) 2005-2006 Micronas USA Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (Version 2) as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/videodev2.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <media/v4l2-subdev.h>
#include <media/v4l2-device.h>

#include "wis-i2c.h"

struct wis_tw2804
{
  struct v4l2_subdev sd;
	int channel;
	int norm;
	int brightness;
	int contrast;
	int saturation;
	int hue;
};

static inline struct wis_tw2804 *to_state(struct v4l2_subdev *sd)
{
  return container_of(sd,struct wis_tw2804,sd);
}

static int tw2804_log_status(struct v4l2_subdev *sd);

static u8 global_registers[]=
{
	0x39, 0x00,
	0x3a, 0xff,
	0x3b, 0x84,
	0x3c, 0x80,
	0x3d, 0x80,
	0x3e, 0x82,
	0x3f, 0x82,
	0xff, 0xff, /* Terminator (reg 0xff does not exist) */
};

static u8 channel_registers[]=
{
	0x01, 0xc4,
	0x02, 0xa5,
	0x03, 0x20,
	0x04, 0xd0,
	0x05, 0x20,
	0x06, 0xd0,
	0x07, 0x88,
	0x08, 0x20,
	0x09, 0x07,
	0x0a, 0xf0,
	0x0b, 0x07,
	0x0c, 0xf0,
	0x0d, 0x40,
	0x0e, 0xd2,
	0x0f, 0x80,
	0x10, 0x80,
	0x11, 0x80,
	0x12, 0x80,
	0x13, 0x1f,
	0x14, 0x00,
	0x15, 0x00,
	0x16, 0x00,
	0x17, 0x00,
	0x18, 0xff,
	0x19, 0xff,
	0x1a, 0xff,
	0x1b, 0xff,
	0x1c, 0xff,
	0x1d, 0xff,
	0x1e, 0xff,
	0x1f, 0xff,
	0x20, 0x07,
	0x21, 0x07,
	0x22, 0x00,
	0x23, 0x91,
	0x24, 0x51,
	0x25, 0x03,
	0x26, 0x00,
	0x27, 0x00,
	0x28, 0x00,
	0x29, 0x00,
	0x2a, 0x00,
	0x2b, 0x00,
	0x2c, 0x00,
	0x2d, 0x00,
	0x2e, 0x00,
	0x2f, 0x00,
	0x30, 0x00,
	0x31, 0x00,
	0x32, 0x00,
	0x33, 0x00,
	0x34, 0x00,
	0x35, 0x00,
	0x36, 0x00,
	0x37, 0x00,
	0xff, 0xff, /* Terminator (reg 0xff does not exist) */
};

static int write_reg(struct i2c_client *client, u8 reg, u8 value, int channel)
{
  int i;
  for(i=0;i<10;i++)
  /*return */if(i2c_smbus_write_byte_data(client,reg|(channel<<6),value)<0)
    return -1;
  return 0;
}

/**static u8 read_reg(struct i2c_client *client, u8 reg, int channel)
{
  return i2c_smbus_read_byte_data(client,reg|(channel<<6));
}*/

static int write_regs(struct i2c_client *client, u8 *regs, int channel)
{
	int i;

	for (i = 0; regs[i] != 0xff; i += 2)
		if (i2c_smbus_write_byte_data(client,
				regs[i] | (channel << 6), regs[i + 1]) < 0)
			return -1;
	return 0;
}

static int wis_tw2804_command(struct i2c_client *client,unsigned int cmd,void *arg)
{
  struct v4l2_subdev *sd=i2c_get_clientdata(client);
  struct wis_tw2804 *dec=to_state(sd);
  int *input;

  printk(KERN_INFO"wis-tw2804: call command %d\n",cmd);
  if(cmd==DECODER_SET_CHANNEL)
  {
    printk(KERN_INFO"wis-tw2804: DecoderSetChannel call command %d\n",cmd);

    input=arg;

    if(*input<0 || *input>3)
    {
      printk(KERN_ERR"wis-tw2804: channel %d is not between 0 and 3!\n",*input);
      return 0;
    }
    dec->channel=*input;
    printk(KERN_DEBUG"wis-tw2804: initializing TW2804 channel %d\n",dec->channel);
    if(dec->channel==0 && write_regs(client,global_registers,0)<0)
    {
      printk(KERN_ERR"wis-tw2804: error initializing TW2804 global registers\n");
      return 0;
    }
    if(write_regs(client,channel_registers,dec->channel)<0)
    {
      printk(KERN_ERR"wis-tw2804: error initializing TW2804 channel %d\n",dec->channel);
      return 0;
    }
    return 0;
  }

  if(dec->channel<0)
  {
    printk(KERN_DEBUG"wis-tw2804: ignoring command %08x until channel number is set\n",cmd);
    return 0;
  }

  return 0;
}

static int tw2804_log_status(struct v4l2_subdev *sd)
{
  struct wis_tw2804 *state=to_state(sd);
  v4l2_info(sd, "Standard: %s\n",state->norm==V4L2_STD_NTSC ? "NTSC" : state->norm==V4L2_STD_PAL ? "PAL" : state->norm==V4L2_STD_SECAM ? "SECAM" : "unknown");
  v4l2_info(sd,"Input: %d\n",state->channel);
  v4l2_info(sd,"Brightness: %d\n",state->brightness);
  v4l2_info(sd,"Contrast: %d\n",state->contrast);
  v4l2_info(sd,"Saturation: %d\n",state->saturation);
  v4l2_info(sd,"Hue: %d\n",state->hue);
  return 0;
}

static int tw2804_queryctrl(struct v4l2_subdev *sd, struct v4l2_queryctrl *query)
{
  static const u32 user_ctrls[]=
  {
    V4L2_CID_USER_CLASS,
    V4L2_CID_BRIGHTNESS,
    V4L2_CID_CONTRAST,
    V4L2_CID_SATURATION,
    V4L2_CID_HUE,
    0
  };

  static const u32 *ctrl_classes[]=
  {
    user_ctrls,
    NULL
  };

  query->id=v4l2_ctrl_next(ctrl_classes,query->id);

  switch (query->id)
  {
    case V4L2_CID_USER_CLASS:
      return v4l2_ctrl_query_fill(query, 0, 0, 0, 0);
    case V4L2_CID_BRIGHTNESS:
      return v4l2_ctrl_query_fill(query, 0, 255, 1, 128);
    case V4L2_CID_CONTRAST:
      return v4l2_ctrl_query_fill(query, 0, 255, 1, 128);
    case V4L2_CID_SATURATION:
      return v4l2_ctrl_query_fill(query, 0, 255, 1, 128);
    case V4L2_CID_HUE:
      return v4l2_ctrl_query_fill(query, 0, 255, 1, 128);
    default:
      return -EINVAL;
  }
}

static int tw2804_s_ctrl(struct v4l2_subdev *sd, struct v4l2_control *ctrl)
{
  struct wis_tw2804 *dec=to_state(sd);
  struct i2c_client *client=v4l2_get_subdevdata(sd);

  int Addr=0x00;
  switch(ctrl->id)
  {
    case V4L2_CID_BRIGHTNESS:
      Addr=0x12;
      break;
    case V4L2_CID_CONTRAST:
      Addr=0x11;
      break;
    case V4L2_CID_SATURATION:
      Addr=0x10;
      break;
    case V4L2_CID_HUE:
      Addr=0x0f;
      break;
    default:
      return -EINVAL;
  }
  ctrl->value=ctrl->value>255 ? 255 : (ctrl->value<0 ? 0 : ctrl->value);
  Addr=write_reg(client,Addr,ctrl->value,dec->channel);
  if(Addr<0)
  {
    printk(KERN_INFO"wis_tw2804: can`t set_ctrl value:id=%d;value=%d",ctrl->id,ctrl->value);
    return Addr;
  }
  printk(KERN_INFO"wis_tw2804: set_ctrl value:id=%d;value=%d",ctrl->id,ctrl->value);
  switch(ctrl->id)
  {
    case V4L2_CID_BRIGHTNESS:
      dec->brightness=ctrl->value;
      break;
    case V4L2_CID_CONTRAST:
      dec->contrast=ctrl->value;
      break;
    case V4L2_CID_SATURATION:
      dec->saturation=ctrl->value;
      break;
    case V4L2_CID_HUE:
      dec->hue=ctrl->value;
      break;
    default:
      return -EINVAL;
  }
  return 0;
}

static int tw2804_g_ctrl(struct v4l2_subdev *sd,struct v4l2_control *ctrl)
{
  struct wis_tw2804 *state=to_state(sd);
 /// struct i2c_client *client=v4l2_get_subdevdata(sd);

  switch(ctrl->id)
  {
    case V4L2_CID_BRIGHTNESS:
      ctrl->value=state->brightness;//=read_reg(client,0x12,state->channel);
      break;
    case V4L2_CID_CONTRAST:
      ctrl->value=state->contrast;//=read_reg(client,0x11,state->channel);
      break;
    case V4L2_CID_SATURATION:
      ctrl->value=state->saturation;//=read_reg(client,0x10,state->channel);
      break;
    case V4L2_CID_HUE:
      ctrl->value=state->hue;//=read_reg(client,0x0f,state->channel);
      break;
    default:
      return -EINVAL;
  }
  return 0;
}

static int tw2804_s_std(struct v4l2_subdev *sd, v4l2_std_id norm)
{
  struct wis_tw2804 *dec=to_state(sd);
  struct i2c_client *client=v4l2_get_subdevdata(sd);

///      v4l2_std_id *input=arg;
  u8 regs[]=
  {
    0x01,norm&V4L2_STD_NTSC ? 0xc4 : 0x84,
    0x09,norm&V4L2_STD_NTSC ? 0x07 : 0x04,
    0x0a,norm&V4L2_STD_NTSC ? 0xf0 : 0x20,
    0x0b,norm&V4L2_STD_NTSC ? 0x07 : 0x04,
    0x0c,norm&V4L2_STD_NTSC ? 0xf0 : 0x20,
    0x0d,norm&V4L2_STD_NTSC ? 0x40 : 0x4a,
    0x16,norm&V4L2_STD_NTSC ? 0x00 : 0x40,
    0x17,norm&V4L2_STD_NTSC ? 0x00 : 0x40,
    0x20,norm&V4L2_STD_NTSC ? 0x07 : 0x0f,
    0x21,norm&V4L2_STD_NTSC ? 0x07 : 0x0f,
    0xff,0xff,
  };
  write_regs(client,regs,dec->channel);
  dec->norm=norm;
  return 0;
}

static const struct v4l2_subdev_core_ops tw2804_core_ops=
{
  .log_status=tw2804_log_status,
  .g_ctrl=tw2804_g_ctrl,
  .s_ctrl=tw2804_s_ctrl,
  .queryctrl=tw2804_queryctrl,
  .s_std=tw2804_s_std,
};

/*
static const struct v4l2_subdev_video_ops tw2804_video_ops = {
  .s_routing = tw2804_s_video_routing,
  .s_fmt = tw2804_s_fmt,
};*/

static const struct v4l2_subdev_ops tw2804_ops=
{
  .core=&tw2804_core_ops,
///  .audio = &s2250_audio_ops,
///  .video = &s2250_video_ops,
};

static int wis_tw2804_probe(struct i2c_client *client,
			    const struct i2c_device_id *id)
{
	struct i2c_adapter *adapter = client->adapter;
	struct wis_tw2804 *dec;
  struct v4l2_subdev *sd;

  printk(KERN_DEBUG "wis_tw2804 :probing %s adapter %s",id->name,client->adapter->name);
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA))
		return -ENODEV;

	dec = kmalloc(sizeof(struct wis_tw2804), GFP_KERNEL);
	if (dec == NULL)
		return -ENOMEM;

  sd=&dec->sd;
	dec->channel = -1;
	dec->norm = V4L2_STD_NTSC;
	dec->brightness = 128;
	dec->contrast = 128;
	dec->saturation = 128;
	dec->hue = 128;
  v4l2_i2c_subdev_init(sd,client,&tw2804_ops);
  v4l2_info(sd,"initializing %s at address 0x%x on %s\n","tw 2804", client->addr, client->adapter->name);

	printk(KERN_DEBUG "wis-tw2804: creating TW2804 at address %d on %s\n",
		client->addr, adapter->name);

	return 0;
}

static int wis_tw2804_remove(struct i2c_client *client)
{
  struct v4l2_subdev *sd=i2c_get_clientdata(client);
  printk(KERN_INFO"wis_tw2804: remove");
  v4l2_device_unregister_subdev(sd);
  kfree(to_state(sd));
	return 0;
}

static const struct i2c_device_id wis_tw2804_id[] = {
	{ "wis_tw2804", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, wis_tw2804_id);

static struct i2c_driver wis_tw2804_driver = {
	.driver = {
		.name	= "WIS TW2804 I2C driver",
	},
	.probe		= wis_tw2804_probe,
	.remove		= wis_tw2804_remove,
	.command	= wis_tw2804_command,
	.id_table	= wis_tw2804_id,
};

static int __init wis_tw2804_init(void)
{
	return i2c_add_driver(&wis_tw2804_driver);
}

static void __exit wis_tw2804_cleanup(void)
{
	i2c_del_driver(&wis_tw2804_driver);
}

module_init(wis_tw2804_init);
module_exit(wis_tw2804_cleanup);

MODULE_LICENSE("GPL v2");
